from abc import abstractmethod, ABCMeta

from pydamain.domain.message import Event


class AbstractEventProducer(metaclass=ABCMeta):
    @abstractmethod
    async def send(self, event: Event):
        ...



from typing import *  # type: ignore


_T_contra = TypeVar("_T_contra", contravariant=True)
_T_co = TypeVar("_T_co", covariant=True)


class _SyncHandler0(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(self, __m: _T_contra, /, **kwargs: Any) -> _T_co:
        ...


class _SyncHandler1(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self, __m: _T_contra, /, __arg_1: Any, **kwargs: Any
    ) -> _T_co:
        ...


class _SyncHandler2(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self, __m: _T_contra, /, __arg_1: Any, __arg_2: Any, **kwargs: Any
    ) -> _T_co:
        ...


class _SyncHandler3(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler4(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler5(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler6(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler7(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler8(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler9(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        __arg_9: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _SyncHandler10(Protocol[_T_contra, _T_co]):
    __name__: str

    def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        __arg_9: Any,
        __arg_10: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


SyncHandler = (
    _SyncHandler0[_T_contra, _T_co]
    | _SyncHandler1[_T_contra, _T_co]
    | _SyncHandler2[_T_contra, _T_co]
    | _SyncHandler3[_T_contra, _T_co]
    | _SyncHandler4[_T_contra, _T_co]
    | _SyncHandler5[_T_contra, _T_co]
    | _SyncHandler6[_T_contra, _T_co]
    | _SyncHandler7[_T_contra, _T_co]
    | _SyncHandler8[_T_contra, _T_co]
    | _SyncHandler9[_T_contra, _T_co]
    | _SyncHandler10[_T_contra, _T_co]
)


class _AsyncHandler0(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(self, __m: _T_contra, /, **kwargs: Any) -> _T_co:
        ...


class _AsyncHandler1(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self, __m: _T_contra, /, __arg_1: Any, **kwargs: Any
    ) -> _T_co:
        ...


class _AsyncHandler2(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self, __m: _T_contra, /, __arg_1: Any, __arg_2: Any, **kwargs: Any
    ) -> _T_co:
        ...


class _AsyncHandler3(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler4(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler5(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler6(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler7(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler8(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler9(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        __arg_9: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


class _AsyncHandler10(Protocol[_T_contra, _T_co]):
    __name__: str

    async def __call__(
        self,
        __m: _T_contra,
        /,
        __arg_1: Any,
        __arg_2: Any,
        __arg_3: Any,
        __arg_4: Any,
        __arg_5: Any,
        __arg_6: Any,
        __arg_7: Any,
        __arg_8: Any,
        __arg_9: Any,
        __arg_10: Any,
        **kwargs: Any,
    ) -> _T_co:
        ...


AsyncHandler = (
    _AsyncHandler0[_T_contra, _T_co]
    | _AsyncHandler1[_T_contra, _T_co]
    | _AsyncHandler2[_T_contra, _T_co]
    | _AsyncHandler3[_T_contra, _T_co]
    | _AsyncHandler4[_T_contra, _T_co]
    | _AsyncHandler5[_T_contra, _T_co]
    | _AsyncHandler6[_T_contra, _T_co]
    | _AsyncHandler7[_T_contra, _T_co]
    | _AsyncHandler8[_T_contra, _T_co]
    | _AsyncHandler9[_T_contra, _T_co]
    | _AsyncHandler10[_T_contra, _T_co]
)


Handler = AsyncHandler[_T_contra, _T_co] | SyncHandler[_T_contra, _T_co]



from typing import ClassVar

from aiokafka import AIOKafkaProducer  # type: ignore

from pydantic import BaseModel

from pydamain.domain.model import Aggregate
from pydamain.domain.message import Event
from pydamain.port.out_.event_producer import AbstractEventProducer


def serialize_key_from_aggregate(aggregate: Aggregate | None):
    if aggregate:
        return "{}: {}".format(
            type(aggregate).__name__, aggregate.id.hex.encode("utf-8")
        )


def serialize_value_from_event(event: Event):
    return event.json(by_alias=True).encode("utf-8")


class KafkaEvent(BaseModel):
    event_type: str
    event: Event


class BaseKafkaEventProducer(AbstractEventProducer):

    host: ClassVar[str]
    port: ClassVar[int]
    topic: ClassVar[str]

    def __init__(self) -> None:
        self.aiokafka_producer = AIOKafkaProducer(
            bootstrap_servers=f"{self.host}:{self.port}",
            key_serializer=serialize_key_from_aggregate,
            value_serializer=serialize_value_from_event,
            compression_type="lz4",
            enable_idempotence=True,  # 멱등성
        )

    def pre_send(self, event: Event):
        ...

    def post_send(self, event: Event):
        ...

    async def send(self, event: Event):
        self.pre_send(event)
        await self.aiokafka_producer.send(  # type: ignore
            topic=self.topic,
            key=event.from_,
            value=KafkaEvent(
                event_type=type(event).__name__,
                event=event,
            ),
        )
        self.post_send(event)



from typing import ClassVar, Literal
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart

from aiosmtplib import SMTP

from pydamain.port.out_.notifications import AbstractEmailNotification


class BaseEmailNotification(AbstractEmailNotification):

    host: ClassVar[str]
    port: ClassVar[int]
    username: ClassVar[str | None] = None
    password: ClassVar[str | None] = None

    def __init__(self) -> None:
        config = {
            "hostname": self.host,
            "port": self.port,
        }
        if self.username and self.password:
            config["username"] = self.username
            config["password"] = self.password
            config["start_tls"] = True
        self.smtp_client = SMTP(**config)

    def pre_send(self):
        ...

    def post_send(self):
        ...

    async def send(
        self,
        from_: str,
        to: str | list[str],
        subject: str,
        text: str,
        subtype: Literal["html", "plain"] = "plain",
        files: list[tuple[str, bytes]] = [],
    ) -> None:
        self.pre_send()
        message: MIMEMultipart = MIMEMultipart("alternative")
        message["From"] = from_
        message["To"] = to
        message["Subject"] = subject
        message.attach(MIMEText(text, subtype, "utf-8"))
        for file_name, file_bytes in files:
            ma = MIMEApplication(file_bytes)
            ma.add_header(
                "content-disposition", "attachment", filename=file_name
            )
            message.attach(ma)
        async with self.smtp_client:
            await self.smtp_client.send_message(message)
        self.post_send()



from contextvars import ContextVar
from datetime import datetime
from typing import ClassVar
from uuid import uuid4
from weakref import WeakValueDictionary
import asyncio

from loguru import logger
from pydantic import BaseModel, Field, UUID4

from pydamain.typing import (
    Any,
    AsyncHandler,
    Handler,
    SyncHandler,
    TypeVar,
    cast,
)
from pydamain.domain.model import Aggregate


class BaseMessage(BaseModel):
    id: UUID4 = Field(default_factory=uuid4)
    create_time: datetime = Field(default_factory=datetime.now)

    class Config:
        frozen = True


R = TypeVar("R")


class Command(BaseMessage):

    handler: ClassVar[Handler["Command", Any] | None] = None

    async def handle(self, deps: dict[str, Any]):
        events = await asyncio.create_task(self._handle(deps))
        await asyncio.gather(
            *(event.handle(deps) for event in events), return_exceptions=True
        )

    async def _handle(self, deps: dict[str, Any]):
        if not self.handler:
            logger.debug(f"{type(self).__name__} is not have handler.")
            return []  # type: ignore
        with EventContext() as event_catcher:
            await handle_message(self, self.handler, deps)
        return event_catcher.events


class Event(BaseMessage):

    from_: Aggregate | None = Field(None, exclude=True)

    handlers: ClassVar[set[Handler["Event", Any]]] = set()

    __event_type_registry__: ClassVar[
        WeakValueDictionary[str, type["Event"]]
    ] = WeakValueDictionary()  # need deserialization from external event

    def __init_subclass__(cls) -> None:
        event_registry[cls.__name__] = cls

    def issue(self):
        event_list = events_context_var.get()
        event_list.append(self)

    async def handle(self, deps: dict[str, Any]):
        events = await asyncio.create_task(self._handle(deps))
        await asyncio.gather(
            *(event.handle(deps) for event in events), return_exceptions=True
        )

    async def _handle(self, deps: dict[str, Any]):
        if not self.handlers:
            logger.debug(f"{type(self).__name__} is not have handlers.")
            return []  # type: ignore
        with EventContext() as event_catcher:
            await asyncio.gather(
                *(
                    handle_message(self, handler, deps)
                    for handler in self.handlers
                ),
                return_exceptions=True,
            )
        return event_catcher.events


event_registry: WeakValueDictionary[str, type["Event"]] = WeakValueDictionary()


def deserialize_external_event(type_: bytes, json: bytes):
    event_cls = event_registry[type_.decode("utf-8")]
    return event_cls.parse_raw(json)


events_context_var: ContextVar[list[Event]] = ContextVar("events_context_var")


class EventContext:
    def __init__(self):
        self.events: list[Event] = []

    def __enter__(self):
        self.token = events_context_var.set(list())
        return self

    def __exit__(self, *args: tuple[Any]):
        self.events.extend(events_context_var.get())
        events_context_var.reset(self.token)


M = TypeVar("M")


async def handle_message(
    msg: M,
    handler: Handler[M, R],
    deps: dict[str, Any],
):
    try:
        logger.debug(f"Handling {msg} with handler {handler.__name__}")
        if asyncio.iscoroutinefunction(handler):
            handler = cast(AsyncHandler[M, R], handler)
            await handler(msg, **deps)
        else:
            handler = cast(SyncHandler[M, R], handler)
            await asyncio.to_thread(handler, msg, **deps)
    except Exception as e:
        logger.exception(f"Exception handling {msg}")
        raise e



from abc import abstractmethod, ABCMeta


class AbstractEmailNotification(metaclass=ABCMeta):
    @abstractmethod
    async def send(self, from_: str, to: str, subject: str, text: str) -> None:
        ...



from __future__ import annotations

from typing import Generic, TypeVar

from abc import ABCMeta, abstractmethod
from uuid import UUID


A = TypeVar("A")


class AbstractRepository(Generic[A], metaclass=ABCMeta):
    @abstractmethod
    def add(self, aggregate: A):
        ...

    @abstractmethod
    async def get(self, identity: UUID) -> A | None:
        ...



from typing import Any, Callable, TypeVar
from uuid import UUID

from sqlmodel.ext.asyncio.session import AsyncSession
from pydamain.domain.model import Aggregate
from pydamain.port.out_.repository import AbstractRepository


A = TypeVar("A", bound=Aggregate)


class BaseSQLModelRepository(AbstractRepository[A]):

    session_factory: Callable[[], AsyncSession]

    def __init_subclass__(cls) -> None:
        if not hasattr(cls, "session_factory"):
            raise TypeError(f"'session_factory' not set to {cls.__name__}.")

    def __init__(self, aggregate_cls: type[A]):
        self._aggregate_cls = aggregate_cls

    async def __aenter__(self):
        self._session = self.session_factory()
        return self

    async def __aexit__(self, *args: tuple[Any]):
        await self._session.rollback()
        await self._session.close()

    def add(self, aggregate: A):
        self._session.add(aggregate)

    async def get(self, identity: UUID) -> A | None:
        return await self._session.get(self._aggregate_cls, identity)

    async def commit(self):
        await self._session.commit()



from typing import Any, ClassVar, cast

from aiokafka import AIOKafkaConsumer  # type: ignore

from pydantic import BaseModel

from pydamain.domain.message import Event, deserialize_external_event
from pydamain.port.in_.event_consumer import AbstractEventConsumer


class KafkaEvent(BaseModel):
    event_type: bytes
    event: bytes


class BaseKafkaEventConsumer(AbstractEventConsumer):

    host: ClassVar[str]
    port: ClassVar[int]
    topic: ClassVar[str]

    def __init__(self, deps: dict[str, Any]) -> None:
        self.aiokafka_consumer = AIOKafkaConsumer(
            self.topic,
            bootstrap_servers=f"{self.host}:{self.port}",
            value_deserializer=KafkaEvent.parse_raw,
            group_id=type(self).__name__,
            enable_auto_commit=False,
            auto_offset_reset="earliest",
        )
        self.deps = deps

    def pre_consume(self, event: Event):
        ...

    def post_consume(self, event: Event):
        ...

    async def consume(self):
        async for msg in self.aiokafka_consumer:  # type: ignore
            msg = cast(KafkaEvent, msg)
            event = deserialize_external_event(
                type_=msg.event_type, json=msg.event
            )
            self.pre_consume(event)
            await event.handle(self.deps)
            await self.aiokafka_consumer.commit()  # type: ignore
            self.post_consume(event)



from __future__ import annotations

from uuid import uuid4
import typing

from sqlmodel import Field, SQLModel, Relationship  # type: ignore
from pydantic import UUID4


class ValueObject(SQLModel):

    id: UUID4 = Field(default_factory=uuid4, primary_key=True, exclude=True)

    class Config(SQLModel.Config):
        validate_assigment = True


class Entity(SQLModel):

    id: UUID4 = Field(default_factory=uuid4, primary_key=True, exclude=True)

    class Config(SQLModel.Config):
        validate_assignment = True

    def __eq__(self, other: typing.Any):
        if isinstance(other, type(self)):
            return self.id == other.id
        return NotImplemented

    def __hash__(self):  # type: ignore
        return hash(self.id)


class Aggregate(Entity):

    version_number: int = Field(..., exclude=True)

    __mapper_args__: dict[str, typing.Any] = {"version_id_col": version_number}

    class Config(Entity.Config):
        ...
